from rest_framework import permissions
from rest_framework.permissions import SAFE_METHODS


class IsStaffOrAdmin(permissions.BasePermission):
    """
    Custom permission to only allow Owners of object or Admin
    TODO : Do check also for Admin user to allow permissions
    """

    def has_object_permission(self, request, view, obj):
        if request.user and (request.user.is_staff or request.user.is_superuser) and request.user.is_active:
            return True
        return False


class IsReadOnly(permissions.BasePermission):
    """
    The request is authenticated as a user, or is a read-only request.
    """

    def has_permission(self, request, view):
        return bool(
            request.method in SAFE_METHODS
        )



from datetime import datetime, timedelta
from humanize import naturaltime


def print_progress_bar(iteration: int, total: int, starting_time=None, decimals=1, length=50, fill='█') -> None:
    """ Call in a loop to create terminal progress bar.

    Shamelessly stolen from: https://stackoverflow.com/questions/3173320/text-progress-bar-in-the-console
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filled_length = int(length * iteration // total)
    bar = fill * filled_length + '-' * (length - filled_length)

    runtime = ''
    if starting_time:
        runtime = naturaltime(datetime.now() - starting_time)

    ending_time = ''
    if starting_time:
        seconds_from_start = (datetime.now() - starting_time).total_seconds()
        per_second = int(iteration / seconds_from_start)
        try:
            ending_time = naturaltime(timedelta(seconds=(-1 * ((total-iteration) / per_second))))
        except (TypeError, ZeroDivisionError):
            ending_time = '???'

    print(
        f'\r{iteration} |{bar}| {percent}% {total} '
        f'[Started: {runtime} Rate: {per_second:d}/per sec Finished: {ending_time}]',
        end='\r')

    # Print New Line on Complete
    if iteration == total:
        print()



import base64
import io
import os
from datetime import timedelta

from PIL import Image

from rest_framework_simplejwt.tokens import Token


class MyToken(Token):
    token_type = 'access'
    lifetime = timedelta(days=1)


def gen_test_user_token(for_user, is_reseller=False, site='example.com'):
    test_token = MyToken.for_user(for_user)
    test_token['site'] = site
    test_token['user_id'] = for_user.username
    test_token['is_reseller'] = is_reseller
    test_token['iss'] = 'nusantara_admin'

    return test_token


def generate_image_file(size, data):
    file = io.BytesIO()
    image = Image.frombytes('L', size, data)
    image.save(file, 'png')
    file.name = 'test.png'
    file.seek(0)
    return file


def generate_base64_image(size: tuple):
    x, y = size
    image = generate_image_file(size, os.urandom(x * y))

    return base64.b64encode(image.read()).decode()



from rest_framework import serializers


class EntityHrefField(serializers.HyperlinkedRelatedField):
    """ Works the same as a hyperlinked related field, just nesting the response in a dict.
    """
    def to_representation(self, value):
        return {
            "title": str(value.name),
            "href": super(EntityHrefField, self).to_representation(value)
        }



import logging
import time
from enum import Enum
from typing import Type

import msgpack
from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site
from django.conf import settings
from django.utils.timezone import now
from kombu import Exchange, producers, Connection, Consumer, Producer, uuid, Queue
from rest_framework.serializers import BaseSerializer

from gramedia.django.amqp_connection import get_publish_connection_and_channel

_logger = logging.getLogger('LOG')
_logger_audit = logging.getLogger('AUDIT')

User = get_user_model()


class EventType(Enum):
    created = 'created'
    changed = 'changed'
    deleted = 'deleted'
    deactivated = 'deactivated'
    activated = 'activated'
    assigned = 'assigned'
    revoked = 'revoked'


class BasicPublisher:
    """ Object used for publishing.
    """

    def __init__(self,
                 exchange_name: str,
                 site: Site,
                 connection: Connection = None,
                 channel=None):
        """ Creates a new instance of BasicPublisher.
        If connection or channel are not defined, they will be initialized
        automatically from the global publish connection and channel objects.

        :param exchange_name:  Name of the exchange this publisher should publish to.
        :param connection: (optional) AMQP connection object.
        :param channel: (optional) AMQP channel that will be used to publish messages.
        """
        self.exchange_name = exchange_name
        self.site = site

        if not all([connection, channel]):
            self._connection, self._channel = get_publish_connection_and_channel()
        else:
            self._connection = connection
            self._channel = channel

    def create_exchange(self, exchange_type='topic') -> None:
        """ Creates an exchange
        """
        if not self._channel.is_open:
            self._connection, self._channel = get_publish_connection_and_channel()
        self._channel.exchange_declare(
            exchange=self.exchange_name,
            type=exchange_type,
            durable=True,
            auto_delete=False
        )

    def get_identity(self, data: dict) -> str:
        return data['href']

    def get_user_href(self, user: User) -> str:
        if isinstance(user, str):
            return user
        return f'https://{self.site.domain}/api/iam/user/{user.username}/' if user else ''

    def simulate_context(self) -> dict:
        """ Simulates a request context sot hat our serializers can operate properly.

        :return: A single-key dictionary containing 'request' with a simulated request object.
        """
        # this is a terrible hack to make the requests still work from the CLI.
        from rest_framework.request import Request, HttpRequest

        class SimulatedHTTPRequest(HttpRequest):
            def _get_scheme(self):
                return "https"

        req = SimulatedHTTPRequest()

        req.META = {
            'SERVER_NAME': self.site.domain,
            'SERVER_PORT': 443,
            'SCRIPT_NAME': settings.FORCE_SCRIPT_NAME
        }
        return {'request': Request(req)}

    def construct_message(self, data: dict, entity_type: str, event_type: EventType, identity: object = None, user: object = None) -> bytes:
        """ Construct the message to be sent.
        :param data: A dictionary of data to be sent
        :param entity_type: Name of the entity
        :param event_type: Event type
        :param identity: Message identifier
        :param user: User that does the event
        :return: A message pack encoded data
        """
        message = {
            "event_time": now().isoformat(),
            "identity": identity or self.get_identity(data),
            "event_type":  event_type.value,
            "entity_type": entity_type,
            "entity_site": self.site.domain,
            "user": self.get_user_href(user) if user else '',
            "data": data
        }
        return msgpack.packb(message, use_bin_type=True)

    def publish(self,
                message: any,
                message_serializer: Type[BaseSerializer],
                event_type: EventType,
                entity_type: str,
                message_identity: any = None,
                user: User = None) -> None:
        """ Publish the message to RabbitMQ.
        :param message:
        :param message_serializer:
        :param event_type:
        :param entity_type:
        :param message_identity:
        :param user:
        """
        self.create_exchange()
        if message_serializer is not None:
            data = message_serializer(message, context=self.simulate_context()).data
        else:
            if type(message) == dict:
                data = message
            else:
                data = {}

        body = self.construct_message(
            data=data,
            entity_type=entity_type,
            event_type=event_type,
            identity=message_identity,
            user=user
        )

        with producers[self._connection].acquire(block=True, timeout=60) as producer:
            the_exchange = Exchange(name=self.exchange_name, type='topic', durable=True, channel=self._channel)
            try:
                _logger.info(f'QUEUE Publish {self.site.domain}-{entity_type}:{event_type} - {data} ')
                producer.publish(
                    body=body,
                    exchange=the_exchange,
                    routing_key=f'{entity_type}.{event_type.value}',
                    declare=[the_exchange],
                    serializer='msgpack',
                    retry=True,
                    retry_policy={
                        'interval_start': 0,  # First retry immediately,
                        'interval_step': 2,   # then increase by 2s for every retry.
                        'interval_max': 30,   # but don't exceed 30s between retries.
                        'max_retries': 30,    # give up after 30 tries.
                    },
                )
            except Exception as exc:
                _logger.exception(f"AMQP error - reconnecting attempt ({exc}) ")


class BasicRpcClient:

    def __init__(self, routing):
        self.connection, self.channel = get_publish_connection_and_channel()
        self.callback_queue = Queue(uuid(), exclusive=True, auto_delete=True)
        self.response = None
        self.routing_key = routing

    def on_response(self, message):
        if message.properties['correlation_id'] == self.correlation_id:
            import msgpack
            self.response = msgpack.unpackb(message.body).get('result')

    def call(self, message: dict, event_type: str, entity_type: str, site: Site) -> any:
        self.response = None
        self.correlation_id = uuid()
        with Producer(self.connection) as producer:
            _logger.info(f'CELERY RPC call {site.domain} with {event_type} - {entity_type} {message} reply to {self.correlation_id}')
            producer.publish(
                {
                    "event_type": event_type,
                    "entity_type": entity_type,
                    "entity_site": site.domain,
                    "data": message
                },
                exchange='',
                routing_key=self.routing_key,
                declare=[self.callback_queue],
                reply_to=self.callback_queue.name,
                correlation_id=self.correlation_id,
                serializer='msgpack',
            )

        with Consumer(self.connection,
                      on_message=self.on_response,
                      queues=[self.callback_queue], no_ack=True):
            _logger.info(f'CELERY RPC call consume {site.domain} with {event_type} - {entity_type} {message} reply to {self.correlation_id}')
            t_current = time.time()
            while self.response is None:
                self.connection.drain_events(timeout=1)
                # time.sleep(0.25)  # sleep for 250 milliseconds
                # if time.time() >= t_current + 60000:
                #     break
        _logger.info(f'CELERY RPC call consume {site.domain} with response {self.response}')

        return self.response



import json
import os

from jsonschema import RefResolver, validators, RefResolutionError
from jsonschema.compat import urlsplit, urldefrag


def find_schema_path(schema_name: str):
    """ Reference resolver for Django based projects.

    This assumes the following:

    - Each 'app' within your django project will have a static/schema directory, and schemas will be stored within
      that directory.
    - Your schemas will not be nested any deeper than <app>/static/schemas/<someschema.json>

    Example:
    # this file exists
    banners/static/schemas/banner-schema.json

    find_schema_path('banner_schema.json')
    > /the-full-path/banners/static/schemas/banner-schema.json

    find_schema_path('i-dont-exist.json')
    > None
    """
    try:
        from django.conf import settings
    except ImportError:
        raise RuntimeError(
            "It doesn't look like you have django installed.  "
            "You cannot use this DjangoSchemaResolver outside of a Django project.")

    for pkg in settings.INSTALLED_APPS:
        full_path = os.path.join(settings.BASE_DIR, *pkg.split('.'), 'static', 'schemas', schema_name)
        if os.path.exists(full_path):
            return full_path


class DjangoSchemaResolver(RefResolver):
    """ Reference resolver for Django based projects.

    This assumes the following:

    - Each 'app' within your django project will have a static/schema directory, and schemas will be stored within
      that directory.
    - Your schemas will not be nested any deeper than <app>/static/schemas/<someschema.json>

    Example:
    banners/static/schemas/banner-schema.json
    """

    def resolve_from_url(self, uri):
        uri, fragment = urldefrag(uri)
        split_url = urlsplit(uri)
        if split_url.scheme in self.handlers:
            return self.handlers[split_url.scheme](uri)
        full_path = find_schema_path(uri.split('/')[-1])
        if full_path:
            with open(full_path) as fp:
                json_schema = json.load(fp)
            if not json_schema:
                # TODO: fix exception method
                raise RefResolutionError(
                    "Unresolvable JSON schema: %r" % uri
                )
            return self.resolve_fragment(json_schema, fragment)
        else:
            # if a schema can't be found locally, just default to using the standard resolver
            return super(DjangoSchemaResolver, self).resolve_from_url(uri)


class DjangoDraft4Validator(validators.Draft4Validator):
    def __init__(self, schema, resolver):
        super(DjangoDraft4Validator, self).__init__(schema, resolver=resolver)



from logging import getLogger

log = getLogger('gramedia')

try:
    import django
    import rest_framework
except ModuleNotFoundError:
    log.exception('This module requires django and rest_framework')
    raise



from django.db import models
from django.db.models import SlugField, Model
from django.utils.text import slugify


def generate_slug(model_class: Model, base_text: str, tries=0) -> str:
    """ Create a unique slug for the given model_class.

    .. warning::

        Makes the slightly naive assumption that your slug field will be called 'slug'

    :param model_class: A django model class (NOT the instance)
    :param base_text: Some text to use as a base for the slug, such as the 'name'
    :param tries: Previous attempts to geenrate the slug
    :return: A unique slug for an item from the database.
    """
    candidate_slug = f'{slugify(base_text)}-{tries}' if tries else slugify(base_text)
    if model_class.objects.filter(slug=candidate_slug).exists():
        return generate_slug(model_class, base_text, tries + 1)
    return candidate_slug


class MonoLangSlugField(SlugField):
    """ Field that automatically generates a slug from a CharField before saving.
    """

    def __init__(self, from_field='name', *args, **kwargs):
        if not 'unique' in kwargs:
            kwargs['unique'] = True
        if not 'max_length' in kwargs:
            kwargs['max_length'] = 150
        super(MonoLangSlugField, self).__init__(*args, **kwargs)
        self.from_field = from_field

    def pre_save(self, model_instance: any, add: bool):
        if not getattr(model_instance, self.attname):
            source = getattr(model_instance, self.from_field)
            setattr(
                model_instance,
                self.attname,
                generate_slug(model_instance.__class__, source)
            )
        return super(MonoLangSlugField, self).pre_save(model_instance, add)


class TimestampedModel(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class BaseModel(TimestampedModel):
    """ Common model for most of our standard use-cases.  Includes a name, slug, and is_active field.
    """
    name = models.CharField(max_length=70)
    slug = MonoLangSlugField()

    is_active = models.BooleanField(default=True)

    class Meta:
        abstract = True

    def __str__(self):
        return self.name



import json
import os

from jsonschema import RefResolver, validators


class CorJsonResolver(RefResolver):
    """ Allows JSON schema to resolve references to other JSONSchemas within our project directory
    """
    def resolve_remote(self, uri):
        _this_dir = os.path.abspath(os.path.dirname(__file__), )
        _app_base_dir = os.path.join(_this_dir, os.pardir)

        try:
            uri = uri.replace(self.base_uri, '')
            uri_parts = uri.split('/')
            package, everything_else = uri_parts[0], uri_parts[1:]

            expected_full_path = os.path.join(
                _app_base_dir,
                package,
                'static',
                'schemas',
                *everything_else)
            if os.path.exists(expected_full_path):
                with open(expected_full_path) as fp:
                    json_schema = json.load(fp)
                return json_schema
        except Exception:
            return super(CorJsonResolver, self).resolve_remote(uri)


class GramediaDraft4Validator(validators.Draft4Validator):
    """  JSON Schema validator.  This is the same as the default validator, for all intents and purposes.
    """
    def __init__(self, schema, resolver):
        super(GramediaDraft4Validator, self).__init__(schema, resolver=resolver)


def validate_gramedia_jsonschema(jsondoc: dict, schema: dict, base_url: str='https://gramedia.com/static/schemas/'):
    """ Validates a JSON schema using GDN's validator.  Where this differs is that it assumes any schemas that start
    with 'base_url' may be located on disk, and will try to locate them based on filename instead of via an HTTP
    url.
    """
    resolver = CorJsonResolver(base_url, jsondoc)
    GramediaDraft4Validator(schema, resolver).validate(jsondoc)



import io
from datetime import timedelta
from http import HTTPStatus
from urllib.parse import quote_plus

import requests
from PIL import Image
from rest_framework_simplejwt.tokens import Token


def snake_to_camel_case(value: str) -> str:
    splited_words = value.split('_')
    return ''.join(splited_words[index] if index == 0 else splited_words[index].title()
                   for index in range(len(splited_words)))


def generate_image_file(size, data):
    file = io.BytesIO()
    image = Image.frombytes('L', size, data)
    image.save(file, 'png')
    file.name = 'test.png'
    file.seek(0)
    return file


class MyToken(Token):
    token_type = 'access'
    lifetime = timedelta(days=1)


def gen_test_user_token(for_user, is_reseller=True, site='example.com'):
    test_token = MyToken.for_user(for_user)
    test_token['site'] = site
    test_token['user_id'] = for_user.username
    test_token['iss'] = 'nusantara_admin'

    return test_token


def is_valid_youtube_url(video_url: str) -> bool:
    try:
        request = requests.head(f'https://www.youtube.com/oembed?format=json&url={quote_plus(video_url)}', timeout=0.5)
        if request.status_code != HTTPStatus.OK:
            return False
    except:
        return False
    return True



"""
Generic HTTP Utilities
======================

Helper classes to make working with HTTP-related headers a little bit easier.
"""
from enum import Enum
import re


class LinkHeaderRel(Enum):
    """ Simple enum that represents some basic 'rel' types that we might
    expect for a 'link header'.

    This is, by no means, comprehensive and is not intended to be.
    """
    first = 'first'
    next = 'next'
    prev = 'prev'
    last = 'last'


class LinkHeaderField(object):
    """ A single entity from a link header field.
    """
    def __init__(self, url: str, rel: LinkHeaderRel, title: str=None):
        self.url = url
        self.rel = rel
        self.title = title

    @classmethod
    def from_string(cls, link_header_part: str):
        return LinkHeaderField(
            url=LinkHeaderField._parse_url(link_header_part),
            rel=LinkHeaderField._parse_rel(link_header_part),
            title=LinkHeaderField._parse_title(link_header_part),
        )

    @staticmethod
    def _parse_rel(link_header_part):
        m = re.search(r';\srel="(?P<rel>\w+)"(;|$)', link_header_part, re.IGNORECASE)
        try:
            rel_str = m.groupdict().get('rel')
            return LinkHeaderRel(rel_str)
        except (ValueError, AttributeError):
            raise MalformedLinkHeader(f"HTTP Link Header missing/malformed 'rel': {link_header_part}")

    @staticmethod
    def _parse_url(link_header_part: str) -> str:
        m = re.search(r'^<(?P<url>.+)>', link_header_part, re.IGNORECASE)
        try:
            return m.groupdict().get('url')
        except AttributeError:
            raise MalformedLinkHeader(f"Invalid format for HTTP Link Header: {link_header_part}.")

    @staticmethod
    def _parse_title(link_header_part: str) -> str:
        m = re.search(r';\stitle="(?P<title>.+)"(;|$)', link_header_part, re.IGNORECASE)
        return m.groupdict().get('title') if m is not None else None

    def __eq__(self, other) -> bool:
        props = ['rel', 'url', 'title', ]
        return all([getattr(self, p) == getattr(other, p) for p in props])

    def __str__(self) -> str:
        link = f'<{self.url}>'
        if self.rel:
            link += f'; rel="{self.rel.value}"' if hasattr(self.rel, 'value') else f'; rel="{self.rel}"'
        link += f'; title="{self.title}"' if self.title else ""
        return link


class LinkHeaderParser(object):
    """ Parses the value of an HTTP Link header, and stores them as a set of
    `LinkHeaderField` objects under the .links attribute.
    """
    def __init__(self, link_header_complete: str):
        """
        :param `str` link_header_complete:
            The complete, unparsed value of an HTTP Link header.
        """
        self.links = [LinkHeaderField.from_string(link.strip()) for link
                      in link_header_complete.split(',')]

    def get(self, link_rel:LinkHeaderRel) -> LinkHeaderField:
        """ Fetches a single `LinkHeaderField`, matching the requested 'rel'

        If no match can be found, returns None.

        :param `LinkHeaderRel` link_rel:
        :return: The link header matching the provided `rel`
        """
        links = [link for link in self.links if link.rel == link_rel]
        return links[0] if links else None


class MalformedLinkHeader(Exception):
    """ Raised when an HTTP link header is in an invalid format.
    """
    pass



from django.conf import settings
from kombu import Connection

_connection = None
_channel = None


def get_publish_connection_and_channel():
    global _channel, _connection
    if not all([_connection, ]):
        _connection = Connection(settings.BROKER_URL)
        _channel = _connection.channel()

    return _connection, _channel



from autoslug import AutoSlugField
from django.db import models
from django.utils.translation import ugettext_lazy as _


class TimestampedModel(models.Model):
    """ An abstract model that contains information about when (and possibly, by whom) an object was last updated.

    Please note, the `modified_by` field is intended to be a human-understandable value to be returned
    by an optimistic-concurrency check failure, and **not** something like a database identifier.
    """
    created = models.DateTimeField(
        _('created'),
        auto_now_add=True,
        help_text=_('Date/time this object was created.')
    )
    modified = models.DateTimeField(
        _('modified'),
        auto_now=True,
        help_text=_('Date/time this object was last updated.')
    )
    modified_by = models.CharField(
        max_length=255,
        blank=True,
        help_text=_('Optionally identifies who last updated the object.')
    )

    class Meta:
        abstract = True


class SoftDeletableModel(models.Model):
    """ A model, which when 'deleted' via an API interface, should be flagged as is_active = false, instead
    of actually being deleted from the database.
    """
    is_active = models.BooleanField(default=True)

    class Meta:
        abstract = True


class ManuallySortedModel(models.Model):
    """ A model, which can have it's display order manually changed via the admin interface
    """
    sort_priority = models.PositiveSmallIntegerField(default=0)

    class Meta:
        abstract = True


class BaseModel(SoftDeletableModel, TimestampedModel):
    """ Common model for most of our standard use-cases.
    """
    name = models.CharField(max_length=120)
    slug = AutoSlugField(populate_from='name', unique=True, max_length=120)

    class Meta:
        abstract = True

    def __str__(self):
        return self.name



from django.db.models import Model
from django.utils.text import slugify


def generate_slug(model_class: Model, base_text: str, tries=0) -> str:
    """ Create a unique slug for the given model_class.

    .. warning::

        Makes the slightly naive assumption that your slug field will be called 'slug'

    :param model_class: A django model class (NOT the instance)
    :param base_text: Some text to use as a base for the slug, such as the 'name'
    :param tries: Previous attempts to geenrate the slug
    :return: A unique slug for an item from the database.
    """
    candidate_slug = f'{slugify(base_text)}-{tries}' if tries else slugify(base_text)
    if model_class.objects.filter(slug=candidate_slug).exists():
        return generate_slug(model_class, base_text, tries+1)
    return candidate_slug



class ReadWriteRouter:
    def db_for_read(self, model, **hints):
        """
        Reads go to a randomly-chosen replica.
        """
        return 'replica'

    def db_for_write(self, model, **hints):
        """
        Writes always go to primary.
        """
        return 'default'

    def allow_relation(self, obj1, obj2, **hints):
        """
        This to allow mismatched using DB
        """
        return True

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        return True



from os.path import dirname, join, abspath
from setuptools import setup, find_packages

ROOT_DIR = dirname(abspath(__file__))

setup(
    packages=find_packages(where=join(ROOT_DIR, 'src')),
    package_dir={'gramedia': 'src/gramedia'},

    project_urls={
        'Documentation': 'https://gdn-python-common.rtfd.io/',
        'Source': 'https://github.com/gramedia-digital-nusantara/python-common',
        'Tracker': 'https://github.com/gramedia-digital-nusantara/python-common/issues',
    },
)



"""
Django Rest Framework utility classes
=====================================

Anything in this package is used along with the Django Rest Framework
to provide for our use cases within our application.
"""
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site
from django.contrib.sites.shortcuts import get_current_site
from django.utils.translation import get_language_from_request
from djangorestframework_camel_case.render import CamelCaseJSONRenderer
from rest_framework import serializers
from rest_framework import pagination
from rest_framework.permissions import BasePermission
from rest_framework.response import Response
from rest_framework.utils.urls import replace_query_param
from rest_framework_simplejwt.authentication import JWTAuthentication

from gramedia.common.http import LinkHeaderField, LinkHeaderRel
from django.utils.translation import ugettext_lazy as _

class LinkHeaderPagination(pagination.PageNumberPagination):
    """ Replaces the default pagination classes, provided by DRF, with one
    that returns pagination data as part of the HTTP Link header.

    The RFC can be found `here<https://www.w3.org/Protocols/9707-link-header.html>`_

    In addition, a second, custom HTTP header is included: X-Total-Results,
    which will be a numeric count of the total number of objects across all
    pages.
    """
    page_size_query_param = 'per_page'
    max_page_size = 250

    def get_first_url(self):
        if not self.page.has_previous():
            return None
        url = self.request.build_absolute_uri()
        return replace_query_param(url, self.page_query_param, 1)

    def get_last_url(self):
        if not self.page.has_next():
            return None
        url = self.request.build_absolute_uri()
        return replace_query_param(
            url, self.page_query_param, self.page.paginator.num_pages)

    def get_paginated_response(self, data):

        next_url = self.get_next_link()
        previous_url = self.get_previous_link()

        link_parts = []

        # if we have a next page, then we know we have a last, as well.
        if next_url is not None:
            link_parts.extend([
                LinkHeaderField(
                    url=next_url,
                    rel=LinkHeaderRel.next,
                    title=str(self.page.number + 1)),
                LinkHeaderField(
                    url=self.get_last_url(),
                    rel=LinkHeaderRel.last,
                    title=self.page.paginator.num_pages),
            ])

        # if we have a previous page, then we know we have a first page, too.
        if previous_url is not None:
            link_parts.extend([
                LinkHeaderField(
                    url=previous_url,
                    rel=LinkHeaderRel.prev,
                    title=str(self.page.number - 1)),
                LinkHeaderField(
                    url=self.get_first_url(),
                    rel=LinkHeaderRel.first,
                    title=str(1)),
            ])

        headers = {
            'X-Total-Results': self.page.paginator.count,
            'X-Page-Size': self.get_page_size(self.request),
            'X-Page': self.page.number
        }

        if link_parts:
            headers['Link'] = ", ".join([str(link) for link in link_parts])

        return Response(data, headers=headers)


def convert_env_boolean(env_value: str) -> bool:
    """ Converts an envvar string into a boolean.  Rules: true/True/TRUE/t/T/1 -> True;  All others false

    >>> convert_env_boolean('t')
    True

    >>> convert_env_boolean('1')
    True

    >>> convert_env_boolean('0')
    False
    """
    return env_value.lower() in ['true', '1', 't']


class SoftDeletableListViewMixin:
    """ Same as a 'ListViewMixin', but instead, by default, filters out
    models which have an 'is_active' attribute set to false.

    If the 'include_inactive=true', parameter is passed in the query string,
    then all models (regardless of 'is_active' attribute), will be returned.
    """
    def get_queryset(self):
        if convert_env_boolean(self.request.query_params.get('include_inactive', '')):
            return self.queryset
        else:
            return self.queryset.filter(is_active=True)


class IsAuthenticatedOrOptions(BasePermission):
    """ Allows access only to authenticated users, unless the request is
    for the http OPTIONS method, then the user is directly allowed.
    """
    def has_permission(self, request, view):
        if request.method == 'OPTIONS':
            return True
        else:
            return request.user and request.user.is_authenticated()


class IsOwner(BasePermission):
    def has_permission(self, request, view):
        if view.queryset.get().user == request.user:
            return True


class SummarizedListMixin:
    def get_serializer_class(self):
        if self.request.method == 'GET':
            if hasattr(self, 'summary_serializer_class'):
                return self.summary_serializer_class
            else:
                return create_summary_serializer(self.serializer_class)
        else:
            return self.serializer_class


def create_summary_serializer(serializer_cls):
    """ Creates a 'summary serializer' for a given standard serializer class.

    This is intended for list endpoints, and shortens the serializer class to only send back a 'title' and an 'href'

    :param serializer_cls:
    :return:
    """
    class SummarySerializer(serializer_cls):
        class Meta(serializer_cls.Meta):
            fields = getattr(
                serializer_cls.Meta,
                'summary_fields',
                ('href', 'name')
            )
    return SummarySerializer


class SummaryCamelCaseRenderer(CamelCaseJSONRenderer):
    """ Just like the CamelCaseJSONRenderer, but replaces any MultiLang fields (or things that
    LOOK like multilang fields) and replaces them with a string representing the user's current language.

    .. code-block:: python

        renderer.render({
            "name": "Derek",
            "bio": {
                "en": "From Pittsburgh",
                "id": "Dari Pittsburgh"
            }
        })

        # this will then return if the request language was set to english
        {
            "name": "Derek",
            "bio": "From Pittsburgh"
        }

    """
    media_type = 'application/json'
    format = 'json'

    def render(self, data, accepted_media_type=None, renderer_context=None):
        # TODO: this is ugly as sin.. needs refactored heavily.
        if isinstance(data, list):
            for element in data:
                enabled_languages = sorted([t[0] for t in settings.LANGUAGES])
                for k, v in element.items():
                    if isinstance(v, dict) and enabled_languages == sorted(v.keys()):
                        preferred_language = settings.LANGUAGE_CODE
                        try:
                            preferred_language = get_language_from_request(renderer_context['request'])
                        except Exception:
                            pass
                        element[k] = v[preferred_language[:2]]  # make if something like en-US, just get 'en'
        elif isinstance(data, dict):
            enabled_languages = sorted([t[0] for t in settings.LANGUAGES])
            for k, v in data.items():
                if isinstance(v, dict) and enabled_languages == sorted(v.keys()):
                    preferred_language = settings.LANGUAGE_CODE
                    try:
                        preferred_language = get_language_from_request(renderer_context['request'])
                    except Exception:
                        pass
                    data[k] = v[preferred_language[:2]]  # make if something like en-US, just get 'en'
        return super(SummaryCamelCaseRenderer, self).render(
            data,
            accepted_media_type=accepted_media_type,
            renderer_context=renderer_context
        )


class FullCamelCaseRenderer(CamelCaseJSONRenderer):
    media_type = 'application/json'
    format = 'json+full'


class CurrentSiteSerializerMixin:
    """ Automatically attaches the current 'site' to a model.

    .. warning::
        This is only for use with FK-Site-related models, with an attribute 'site'
    """

    def create(self, validated_data):
        validated_data['site'] = get_current_site(self.context['request'])
        return super().create(validated_data)

    def update(self, instance, validated_data):
        validated_data['site'] = get_current_site(self.context['request'])
        return super().update(instance, validated_data)


class CurrentSiteViewSetMixin:

    def get_queryset(self):
        site = get_current_site(self.request)
        return self.queryset.filter(site=site)


class JWTNusantaraAuthentication(JWTAuthentication):
    request = None

    def authenticate(self, request):
        header = self.get_header(request)

        if header is None:
            return None

        self.request = request

        raw_token = self.get_raw_token(header)
        if raw_token is None:
            return None

        validated_token = self.get_validated_token(raw_token)

        return self.get_user(validated_token), validated_token

    def get_user(self, validated_token):
        """
        Attempts to find and return a user using the given validated token.
        """
        from rest_framework_simplejwt.exceptions import AuthenticationFailed
        from rest_framework_simplejwt.exceptions import InvalidToken
        from rest_framework_simplejwt.settings import api_settings

        try:
            user_id = validated_token[api_settings.USER_ID_CLAIM]
        except KeyError:
            raise InvalidToken(_('Token contained no recognizable user identification'))

        try:
            user_site = validated_token['site']
        except KeyError:
            raise InvalidToken(_('Token contained no recognizable site identification'))

        site = None
        if self.request:
            site = Site.objects.get_current(self.request)
        if site is None:
            raise AuthenticationFailed(_('Site not found'), code='user_not_found')
        if user_site != site.domain:
            raise AuthenticationFailed(_('Site is not match'), code='user_not_found')

        try:
            user = get_user_model().objects.get(**{api_settings.USER_ID_FIELD: user_id})
        except get_user_model().DoesNotExist:
            raise AuthenticationFailed(_('User not found'), code='user_not_found')

        if not user.is_active:
            raise AuthenticationFailed(_('User is inactive'), code='user_inactive')

        return user

def is_normal_user(request):
    if not request.user:
        return True

    if not(request.user.is_staff or request.user.is_superuser):
        return True

def get_entity_href_serializer(model_class, meta_extra_kwargs=None, *init_args, **init_kwargs):
    class EntityHrefSerializer(serializers.HyperlinkedModelSerializer):
        name = serializers.CharField(required=False)

        class Meta:
            model = model_class
            fields = ('href', 'name',)
            # extra_kwargs = meta_extra_kwargs or {'href': {'lookup_field': 'slug', }, }

    return EntityHrefSerializer(*init_args, **init_kwargs)



"""
Config Helpers
==============

Simple helper-methods meant for parsing application config settings from
environmental variables.  The intention of this module is to make 'config.py'
more readable.
"""
import re
from os import getenv
from functools import partial
from typing import Callable


def __env(app_prefix: str, var_name: str, default: any=None, var_type: Callable[[str], any]=str) -> any:
    """ Fetches an environmental variable and returns it's value as type.
    """
    value = getenv(f'{app_prefix}_{var_name}')

    if value is None:
        return default

    return var_type(value)


_str = partial(__env, var_type=str)


def _bool(app_prefix: str, var_name: str, default: bool=None) -> bool:
    """ Converts an environmental variable to a boolean.
    Accepted truthy values: 1, true, t, y, yes (similar to YAML, with the addition of 1)
    This converter is case-insensitive.
    """
    truthy_values = ['1', 'true', 't', 'y', 'yes', ]
    return __env(app_prefix, var_name, default, lambda x: x.lower() in truthy_values)


def _int(app_prefix: str, var_name: str, default: int=None) -> int:
    """ Converts an environmental variable to an integer.
    """
    return __env(app_prefix, var_name, default, int)


class BadDatabaseUri(Exception):
    pass


def _db_url_to_django(app_prefix: str, var_name: str, default: dict=None):
    """ Converts a database URI string (similar to SQLAlchemy format) to Django.

    **Examples**

    +--------------------------+-----------------------------------------------------------------------+
    | Backend                  | Format                                                                |
    +==========================+=======================================================================+
    | Postgresql/MySQL/Oracle  | {postgresql|mysql|oracle}://{user}:{password}@{host}:{port}/{db_name} |
    | SQlite                   | sqlite3://{absolute_path}                                             |
    +--------------------------+-----------------------------------------------------------------------+
    """
    val = __env(app_prefix, var_name, default, str)

    protocol, rest = re.match(r'^(\w+)://(.*)$', val).groups()
    if not all([protocol, rest]):
        raise BadDatabaseUri()

    if protocol == 'sqlite3':
        regex = re.compile(r'^(?P<NAME>.*)$')
    else:
        regex = re.compile(r'^(?P<USER>\w+?):(?P<PASSWORD>.+)@(?P<HOST>\S+?):(?P<PORT>\d+)/(?P<NAME>\w+)$')

    try:
        rslt = regex.match(rest).groupdict()
        rslt['ENGINE'] = f'django.db.backends.{protocol}'

    except AttributeError:
        raise BadDatabaseUri()

    return rslt


class EnvConfig:
    """ Class that should be used when referencing environmental variables.

    It supports a prefix, so that all environmental variables will automatically be specified
    with some prefix.

    >>> import os
    >>> os.environ.setdefault('GM_DB_PORT', '5432')
    '5432'
    >>> env = EnvConfig('GM')
    >>> env.int('DB_PORT')
    5432

    >>> os.environ.setdefault('GM_DEBUG', 'yes')
    'yes'
    >>> env.boolean('DEBUG')
    True
    """
    def __init__(self, app_prefix: str):
        self.app_prefix = app_prefix

    def boolean(self, var_name: str, default: bool=None) -> bool:
        return _bool(self.app_prefix, var_name, default)

    def string(self, var_name: str, default: str=None) -> str:
        return _str(self.app_prefix, var_name, default)

    def int(self, var_name: str, default: int=None) -> int:
        return _int(self.app_prefix, var_name, default)

    def django_db(self, var_name: str, default: str=None) -> dict:
        return _db_url_to_django(self.app_prefix, var_name, default)



import logging
import os
from urllib.parse import urlparse

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.sites.models import Site

from gramedia.django.signalling import EventType

_logger = logging.getLogger('LOG')
_logger_audit = logging.getLogger('AUDIT')

UserModel = get_user_model()

_LOGGER_KEY = 'SECAGUSER'

__ALL__ = ['process_security_auth_group_message', ]


def process_security_auth_group_message(message):
    _logger.info(f'[*] {_LOGGER_KEY} Consuming Security Auth Group - User queue')
    event_type = message.get('event_type', '')
    entity = message.get('data')

    identity_parts = urlparse(message.get('identity'))
    site = Site.objects.get(domain=message.get('entity_site') or identity_parts.hostname)
    customer_identity_parts = urlparse(message.get('customer'))
    customer_identity = os.path.basename(os.path.normpath(identity_parts.path))
    try:
        user = get_user_model().objects.filter(username=customer_identity).first()
        if not user:
            raise UserModel.DoesNotExist

        access_group = Group.objects.get(id=entity.get('access_group'))
    except Group.DoesNotExist as exc:
        _logger.exception(f'{_LOGGER_KEY} not find access group {entity.get("access_group")}')
        raise
    except UserModel.DoesNotExist:
        _logger.exception(f'{_LOGGER_KEY} not find user {entity.get("customer")}')
        raise

    _logger.info(f"{_LOGGER_KEY} Receive business {message}")
    if user and access_group:
        if event_type == EventType.assigned.value:
            access_group.user_set.add(user)
        elif event_type == EventType.revoked.value:
            access_group.user_set.remove(user)
        else:
            _logger.error(f'{_LOGGER_KEY} Cannot match the event {event_type}')
    else:
        _logger.error(f'{_LOGGER_KEY} Cannot match the user or auth group')
